#include <stdlib.h> /* for rand() */
#include <string.h> /* for memset() */

#if !defined(__cplusplus)
typedef int bool;
#define false   0
#define true    1
#endif

#define HOOP    10
#define GOAL   100
#define MAXPOS 109

typedef struct Board Board;
struct Board {
    int   b_whosemove; /* player number, 0 or 1 */
    int   b_prevmove;  /* the forbidden move, if any */
    int   b_where[2];  /* current position */
    int   b_inhoop[2]; /* 0 if ok, +n if mid-hoop, -1 if overshot */
    bool  b_missed[2]; /* true if player has already overshot once */
};


static void makemove(Board *b, int mv) {
    int dist;
    int iw = b->b_whosemove;
    int amv = (mv >= 0 ? mv : -mv);
    dist = (b->b_where[iw] + mv) % HOOP;
    if (dist > 0 && mv > dist && mv != 2*dist) b->b_missed[iw] = true;
    b->b_where[iw] += mv;
    if (mv == 0) {
	/* no change */
    } else if (mv <= dist) {
	b->b_inhoop[iw] = 0;
    } else if (dist == 0) {
	b->b_inhoop[iw] = mv;
    } else if (mv == dist * 2) {
	b->b_inhoop[iw] = 0;
    } else {
	b->b_inhoop[iw] = -1;
    }
    b->b_prevmove = (b->b_inhoop[iw] > 0 ? 0 : amv);
    b->b_whosemove = (iw + 1) % 2;
}

/*
 * Generate all possible plays.  The caller must supply a sufficiently
 * large array; 8 is plenty.
 */
static int genmoves(int *moves, Board const *b) {
#define BADMOVE(m) ((m) == b->b_prevmove || (m) == 9 - b->b_prevmove)
    int m;
    int iw = b->b_whosemove;
    int nmoves = 0;
    if (b->b_inhoop[iw] == 0) {
	for (m = 1; m <= 8; ++m) {
	    if (b->b_missed[iw]) {
		int dist = (b->b_where[iw] + m) % HOOP;
		if (dist > 0 && m > dist && m != 2*dist) continue;
	    }
	    if (!BADMOVE(m)) moves[nmoves++] = m;
	}
    } else if (b->b_inhoop[iw] > 0) {
	moves[nmoves++] = (BADMOVE(b->b_inhoop[iw]) ? 0 : b->b_inhoop[iw]);
    } else {
	for (m = -(b->b_where[iw] % HOOP + 1); m >= -8; --m) {
	    if (!BADMOVE(-m)) moves[nmoves++] = m;
	}
    }
    if (nmoves == 0) abort();
    return (nmoves);
#undef BADMOVE
}


/* actually a small integer 1=lose 2=bad 3=draw 4=good 5=win */
/* also 0=unknown, in some circumstances */
typedef unsigned char Outcome;
#define ORIGIN 3
#define RTABSIZ (32*10*3*10*10*sizeof(Outcome))
static Outcome (*rtable)[10][3][10][10];
static bool didinit = false;

static void inittable(void) {
    didinit = true;
    rtable = (Outcome (*)[10][3][10][10])malloc(RTABSIZ);
    memset(rtable, 0, RTABSIZ);
}

/* test whether we're in one of the known deadly positions, given k > 0 */
static bool cascade(Board const *bp) {
    int t = bp->b_where[bp->b_whosemove] % HOOP;
    int k = bp->b_inhoop[1-bp->b_whosemove];
    if (k > 4) k = 9 - k;
    if (k == 1 && t < 7 && (bp->b_inhoop[bp->b_whosemove]&~1) == 0) return (true);
    if (k == 2 && ((t&1) != 0 || t < 4) && (bp->b_inhoop[bp->b_whosemove]&~2) == 0) return (true);
    return (false);
}

/*
 * If we already know the result, return it.
 * If it would require some nontrivial computation, then just return 0.
 * Actually return a pointer so that the caller can update it with the
 * correct value.
 */
static Outcome *fastsolve(Board const *bp) {
    Outcome *r;
    int d0, d1;
    if (bp->b_inhoop[0] == 0 && bp->b_inhoop[1] == 0 && (d0=bp->b_where[0]/HOOP) < GOAL/HOOP && (d1=bp->b_where[1]/HOOP) < GOAL/HOOP && (unsigned)(1+d0-d1)<3) {
	int ix = bp->b_prevmove < 5 ? bp->b_prevmove-1 : 8 - bp->b_prevmove;
	if (bp->b_whosemove != 0) ix |= 4;
	if (bp->b_missed[0]) ix |= 8;
	if (bp->b_missed[1]) ix |= 16;
	r = &rtable[ix][d0][1+d0-d1][bp->b_where[0]%HOOP][bp->b_where[1]%HOOP];
	if ((unsigned)((char *)r - (char *)rtable) >= RTABSIZ) abort();
    } else {
	static Outcome scratch;
	if (bp->b_where[1] > GOAL && bp->b_inhoop[1] == 0) {
	    if (bp->b_where[0] > GOAL && bp->b_inhoop[0] == 0) {
		scratch = ORIGIN;
	    } else {
		scratch = (bp->b_whosemove == 1 ? ORIGIN+2 : ORIGIN-2);
	    }
	} else if (bp->b_whosemove == 0 && bp->b_where[0] > GOAL && bp->b_inhoop[0] == 0) {
	    scratch = ORIGIN+2;
	} else if (bp->b_inhoop[1-bp->b_whosemove] > 0 && cascade(bp)) {
	    scratch = ORIGIN+2;
	} else if (bp->b_inhoop[0] == 0 && bp->b_inhoop[1] == 0 && bp->b_where[bp->b_whosemove]/HOOP >= bp->b_where[1-bp->b_whosemove]/HOOP + 2) {
	    scratch = ORIGIN+1;
	} else {
	    scratch = 0;
	}
	r = &scratch;
    }
    return (r);
}

static Outcome solve(Board const *bp) {
    int mp[8];
    int nm;
    int i;
    int rc;
    int t;
    Outcome *r = fastsolve(bp);
    if (*r != 0) return (*r);
    nm = genmoves(mp, bp);
    t  = bp->b_where[bp->b_whosemove] % HOOP;
    if (t != 0) {
	static int const goodmove[] = { 0, 6, 7, 5, 4, 1, 8, 6, 4, 2 };
	static int const fairmove[] = { 0, 0, 5, 0, 3, 2, 0, 2, 0, 0 };
	int const g = goodmove[t];
	int const f = fairmove[t];
	int j = 0;
	for (i = 0; i < nm; ++i) {
	    int m = mp[i];
	    if (j == 1 && m == g) {
		mp[0] = m;  mp[i] = f;  m = f;
	    }
	    if (m == f || m == g) {
		mp[i] = mp[j];  mp[j] = m;
		++j;
		if (f == 0 || j == 2) break;
	    }
	}
    }
    rc = ORIGIN-2;
    for (i = 0; i < nm && rc < ORIGIN+1; ++i) {
	Board newb;
	int tc;
	newb = *bp;
	makemove(&newb, mp[i]);
	if ((tc = *fastsolve(&newb)) != 0) {
	    tc = 2*ORIGIN - tc;
	    if (tc > rc) rc = tc;
	}
    }
    for (i = 0; i < nm && rc < ORIGIN+1; ++i) {
	Board newb;
	int tc;
	newb = *bp;
	makemove(&newb, mp[i]);
	tc = 2*ORIGIN - solve(&newb);
	if (tc > rc) rc = tc;
    }
    *r = rc;
    return (rc);
}

static int pickmove(Board const *bp, int const *mp, int nm) {
    int i;
    int besti[8];
    int nbest;
    Outcome bestval;
    /* Now find the best move... */
    bestval = ORIGIN+3;
    nbest = 0;
    for (i = 0; i < nm; ++i) {
	Board newb;
	Outcome val;
	newb = *bp;
	makemove(&newb, mp[i]);
	val = solve(&newb);
	if (val <= bestval) {
	    if (val < bestval) nbest = 0;
	    besti[nbest++] = i;
	    bestval = val;
	}
    }
    nm = nbest;
    if (nm == 0) abort();
    i = rand() % nm;
    return (mp[besti[i]]);
}


static Board bk;

int move1(int score1, int score2, int lastmove) {
    int movebuf[8];
    int nm, mv;
    if (!didinit) inittable();
    if (bk.b_whosemove != 0) makemove(&bk, score2 - bk.b_where[1]);
    if (bk.b_where[0] == 0) {
	mv = 2;
    } else {
	nm = genmoves(movebuf, &bk);
	mv = (nm == 1 ? movebuf[0] : pickmove(&bk, movebuf, nm));
    }
    makemove(&bk, mv);
    return (mv);
}

int move2(int score1, int score2, int lastmove) {
    int movebuf[8];
    int nm, mv;
    if (!didinit) inittable();
    if (bk.b_whosemove != 1) makemove(&bk, score1 - bk.b_where[0]);
    if (bk.b_where[1] == 0) {
	mv = (bk.b_where[0] == 2 ? 8 : bk.b_where[0]/2 == 3 ? 13-bk.b_where[0] : 2);
    } else {
	nm = genmoves(movebuf, &bk);
	mv = (nm == 1 ? movebuf[0] : pickmove(&bk, movebuf, nm));
    }
    makemove(&bk, mv);
    return (mv);
}

